package ap.proof.goal;

import ap.terfor.{ArithConj, Conjunction, Formula, ConstantTerm, VariableTerm,
                  TermOrder, TerFor, Term, Quantifier}
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.substitutions.ConstantSubst
import ap.util.{Debug, FilterIt, Logic}

object Goal {
  
  private val AC = Debug.AC_GOAL

  def apply(facts : Conjunction, tasks : TaskManager, age : Int,
            eliminatedConstants : Set[ConstantTerm], order : TermOrder) : Goal = {
    if (facts.isFalse) {
      TRUE
    } else {
      new Goal(facts, tasks, age, eliminatedConstants, order)
    }
  }
  
  def apply(f : Formula, order : TermOrder) : Goal =
    apply(Conjunction.TRUE,
          TaskManager.EMPTY + new ComplexFormulaTask (f),
          0,
          Set.empty,
          order)
  
  def apply(f : Formula,
            eliminatedConstants : Set[ConstantTerm],
            order : TermOrder) : Goal =
    apply(Conjunction.TRUE,
          TaskManager.EMPTY + new ComplexFormulaTask (f),
          0,
          eliminatedConstants,
          order)
  
  val TRUE : Goal = new Goal (Conjunction.FALSE, TaskManager.EMPTY, 0,
                              Set.empty, TermOrder.EMPTY)
  
}

class Goal private (val facts : Conjunction,
                    val tasks : TaskManager,
                    val age : Int,
                    val eliminatedConstants : Set[ConstantTerm],
                    val order : TermOrder) extends ProofTree {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(Goal.AC,
                   // facts are really just literals (equations, inequalities,
                   // predicate literals)
                   facts.quans.isEmpty && facts.negatedConjs.isTrue &&
                   // there should not be any free variables (use constants
                   // instead)
                   facts.variables.isEmpty &&
                   // if a contradiction is detected, we can forget about
                   // everything
                   (!facts.isFalse || tasks.isEmpty) &&
                   // we must choose an ordering such that the constants to be
                   // eliminated are maximal
                   (facts isSortedBy order) &&
                   (order constantsAreMaximal eliminatedConstants))
  //////////////////////////////////////////////////////////////////////////////
  
  private def elimConstants(tfs : Iterator[LinearCombination])
                                                : Iterator[LinearCombination] =
    FilterIt(tfs, (tf:LinearCombination) => {
      val res = tf.isEmpty || !(this eliminates tf.leadingTerm)
      //////////////////////////////////////////////////////////////////////////
      Debug.assertPost(Goal.AC,
                       res == (tf.constants intersect eliminatedConstants).isEmpty)
      //////////////////////////////////////////////////////////////////////////
      res
    })
    
  /**
   * Check whether the given left-hand side of an equation in the antecedent
   * encodes a divisibility condition
   */
  private def isDivisibilityEq(lc : LinearCombination) : Boolean = {
    val res = !lc.isEmpty &&
              (this eliminates lc.leadingTerm) &&
              !lc.leadingCoeff.isOne &&
              // check that all other eliminable constants have been eliminated
              (lc.size <= 1 || !(this eliminates (lc(1) _2)))
    
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPost(Goal.AC,
                     !res || Logic.forall(1, lc.size, (i) =>
                                          !(this eliminates (lc(i) _2))))
    ////////////////////////////////////////////////////////////////////////////
    
    res
  }
    
  /**
   * Extract the divisibility conditions that are encoded in the equations of
   * the antecedent. An equation <code>n*c + t = 0</code> becomes the condition
   * <code>EX (n*_0 + t = 0)</code>
   */
  private def divisibilityConds : Formula = Conjunction.conj(
    for (lc <- FilterIt(facts.arithConj.positiveEqs.elements,
                        isDivisibilityEq _)) yield {
      //////////////////////////////////////////////////////////////////////////
      Debug.assertInt(Goal.AC,
                      lc.leadingTerm.isInstanceOf[ConstantTerm] &&
                      lc.variables.isEmpty)
      //////////////////////////////////////////////////////////////////////////
      
      val subst = new ConstantSubst(lc.leadingTerm.asInstanceOf[ConstantTerm],
                                    VariableTerm(0), order)
                                    
      Conjunction.quantify(Array(Quantifier.EX),
                           EquationConj(subst(lc), order),
                           order)
    } , order)
    
  lazy val closingConstraint : Formula = {
    val ac = facts.arithConj
    
    Conjunction.conj(Array(
      ArithConj(EquationConj(elimConstants(ac.positiveEqs.elements), order),
                NegEquationConj(elimConstants(ac.negativeEqs.elements), order),
                order),
      divisibilityConds),
      order).negate
  }
  
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Return whether <code>t</code> is a constant that is eliminated in this goal
   */
  def eliminates(t : Term) : Boolean = t match {
    case t : ConstantTerm => eliminatedConstants contains t
    case _ => false
  }

  //////////////////////////////////////////////////////////////////////////////

  /**
   * Update the goal after performing some task
   */
  def taskFinished(updatedFacts : Conjunction,
                   updatedElimConstants : Set[ConstantTerm],
                   updatedOrder : TermOrder,
                   newTasks : Iterable[Task],
                   taskFinishedAction : TaskFinishedAction) : ProofTree = {
    val remTasks =
      if (taskFinishedAction.removeFirstTask) tasks.removeFirst else tasks
    val updatedTasks = remTasks ++ newTasks
    
    Goal(updatedFacts, updatedTasks, age + 1,
         updatedElimConstants, updatedOrder)
      .continueAfterTaskFinished(taskFinishedAction)
  }

  def taskFinished(updatedElimConstants : Set[ConstantTerm],
                   updatedOrder : TermOrder,
                   newTasks : Iterable[Task],
                   taskFinishedAction : TaskFinishedAction) : ProofTree =
    taskFinished(facts, updatedElimConstants, updatedOrder, newTasks,
                 taskFinishedAction)

  def taskFinished(taskFinishedAction : TaskFinishedAction) : ProofTree =
    taskFinished(facts, eliminatedConstants, order, Array(), taskFinishedAction)

  def taskFinished(updatedFacts : Conjunction,
                   newTasks : Iterable[Task],
                   taskFinishedAction : TaskFinishedAction) : ProofTree =
    taskFinished(updatedFacts, eliminatedConstants, order, newTasks, taskFinishedAction)

  def taskFinished(updatedFacts : Conjunction,
                   taskFinishedAction : TaskFinishedAction) : ProofTree =
    taskFinished(updatedFacts, eliminatedConstants, order, Array(), taskFinishedAction)

  def taskFinished(newTasks : Iterable[Task],
                   taskFinishedAction : TaskFinishedAction) : ProofTree =
    taskFinished(facts, eliminatedConstants, order, newTasks, taskFinishedAction)

  //////////////////////////////////////////////////////////////////////////////
      
  def stepPossible : Boolean = !tasks.isEmpty

  def step : ProofTree = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(Goal.AC, this.stepPossible)
    ////////////////////////////////////////////////////////////////////////////
    
    // apply the first task to the goal and then stop immediately
    tasks.max(this, TaskFinishedAction.STOP_AND_REMOVE)
  }

  def runRunRun : ProofTree = {
    if (stepPossible) {
      // apply the first task to the goal and then continue
      tasks.max(this, TaskFinishedAction.CONT_AND_REMOVE)
    } else {
      println("No more rules: ")
      println(this)
      this
    }
  }

  /**
   * Decide how to continue after performing one task
   */
  private def continueAfterTaskFinished(taskFinishedAction : TaskFinishedAction)
                                                              : ProofTree = {
    if (taskFinishedAction.contWithTasks)
      runRunRun
    else
      this
  }
  
  //////////////////////////////////////////////////////////////////////////////
    
  override def toString : String = {
    "Facts: " + facts + "\n" +
    "Tasks: " + tasks.length + "\n" +
    "Eliminated constants: " + eliminatedConstants + "\n" +
    "Constraint: " + closingConstraint
  }
}
